program clock ;

  const

    acia = $f7e8 ; { terminal uart location }
    
  { task identification
    task identifications are bytes
    task identification of zero and #255 are not allowed
    the fall thru task in main program is assigned task id = 1
    each task's id must be unique }

    set_alarm_id = #7 ;
    set_time_id = #6 ;
    command_id = #5 ;
    display_time_id = #4 ;
    display_alarm_id = #3 ;
    update_time_id = #2 ;
    main_id = #1 ;

  { task priority
    task prioritys are non-zero bytes (#1 .. #255)
    the smaller the number the higher the task's priority

  { task timeslice
    task timeslices are bytes
    task timeslice of zero is not allowed  }
    task_timeslice = #10 ; { 10 ms }

  { total number of tasks to run }
    total_tasks = #7 ;

  { the data area is split into system and data stacks
    by the activate routines. the data area size is the
    total size }
    task_data_size = 200 ; {120 for data}
    task_system_size = 80 ;

  type

{$Itcb_nb}

  var

  { global task events }

    abort_event,
    time_event,
    alarm_event   : ^tcb ;

  { global task regions }

    crt_region,
    time_region   : ^tcb ;

  { global task variables }

    count         : byte ;
    task_sel      : byte ;
    
    alarm_hour,
    alarm_minute,
    alarm_second  : byte ;

    hours,
    minutes,
    seconds       : byte ;
    
  { tcbtable }

    tcbtab        : array [#0 .. total_tasks] of tcb ;

  { data areas for tasks }

    task_stacks : array [#2 .. total_tasks] of
                  array [1 .. task_data_size] of byte ;
    
    term : record
             stat, data : byte
           end                  at acia ;
    
{$Iproc_nb}

  procedure setint ; external ;
  
  procedure clearint ; external ;

  procedure write (ch : char) ;
    begin
      while (term.stat and #%00000010) = #0 do ;
      term.data := ch
    end ;
    
  procedure writeln (s : string[30]) ;
    var
      x : byte ;
    begin
      while not enter(crt_region) do ;
      for x := #1 to s[#0] do
        write (s[x]) ;
      write (#$d) ;
      write (#$a) ;
      leave(crt_region)
    end ;

  {   
    write the current time to crt
    note : the calling procedure must enter and leave crt_region
  }
  procedure write_time (hour      : byte ;
                        minute    : byte ;
                        second    : byte) ;
    begin
      writeln (concat(substr(string(ord(hour)+100),2,2),':',
                      substr(string(ord(minute)+100),2,2),':',
                      substr(string(ord(second)+100),2,2)))
    end ; { write_time }

  procedure read (var ch : char) ;
    begin
      while not odd (term.stat) do 
        delay (10) ;
      ch := term.data and #$7f
    end ;
    
  procedure readdigits (var digits : byte) ;
    var
      c : char ;
      b : byte ;
    begin
      digits := #0 ;
      for b := #1 to #2 do
        begin
          repeat
            read (c) ;
            write (c)
          until c in ['0'..'9'] ;
          digits := digits * #10 + c - '0'
        end
    end ;
      
  procedure readtime (var hr, min, sec : byte) ;
    begin
      while not enter(crt_region) do ;
      readdigits (hr) ;
      readdigits (min) ;
      readdigits (sec) ;
      leave (crt_region) ;
      writeln ('')
    end ;
    
  procedure set_alarm ;
    begin
      repeat
        writeln ('') ;
        writeln ('Enter alarm time HH MM SS') ;
        readtime (alarm_hour, alarm_minute, alarm_second) ;
        unblock (command_id) ;
        block (set_alarm_id)
      until false ;
    end ; { set_alarm }

  procedure set_time ;
    begin
      repeat
        writeln ('') ;
        writeln ('Enter new time HH MM SS') ;
        while not enter (time_region) do ;
        readtime (hours, minutes, seconds) ;
        leave (time_region) ;
        unblock (command_id) ;
        block (set_time_id)
      until false ;
    end ; { set_time }

  procedure command ;
    var
      cmd : char ;
    begin
      repeat
        read (cmd) ;
        cmd := upshift(cmd) ;
        case cmd of
        'Q' : cause (abort_event) ;
        'T' : writeln ('testing') ;
        'S' : 
          begin
            unblock (set_time_id) ;
            block (command_id)
          end ;
        'A' : 
          begin
            unblock (set_alarm_id) ;
            block (command_id)
          end
        end
      until false ;
    end ; { command }

  {
    task 4 will wait for 1 time event and
    then output to the crt the current time
    note : the crt screen will be cleared and
           the header outputted only once
  }
  procedure display_time ;
    var
      count : byte ;
    begin
      writeln ('hh:mm:ss') ;
      repeat
        count := #1 ;
        wait (time_event,count,0) ;
        write_time (hours,minutes,seconds)
      until false ;
    end ; { display_time }

  {
    task 3 will wait for 1 alarm event and
    then output to the crt the alarm time
  }
  procedure display_alarm ;
    var
      count : byte ;
    begin
      repeat
        count := #1 ;
        wait (alarm_event,count,0) ;
        writeln (' alarm at ') ;
        write_time (hours,minutes,seconds) ;
        writeln (#7) { bell }
      until false ;
    end ; { display_alarm }

  {
    task 2 will delay for one second and
    then update the global time after that
    it will generate a time event
  }
  procedure update_time ;
    const
      second_timeout = 1000 ; { 1000 ticks * 1 ms per tick }
    var
      abort, alarm : boolean ;
    begin
      repeat
        delay (second_timeout) ;
        while not enter(time_region) do ;
        seconds := seconds + #1 ;
        if seconds = #60
          then 
            begin
              seconds := #0 ;
              minutes := minutes + #1 ;
              if minutes = #60
                then 
                  begin
                    minutes := #0 ;
                    hours := hours + #1 ;
                    if hours = #24
                      then 
                        hours := #0 ;
                  end ; { if minutes }
            end ; { if seconds }
        leave (time_region) ;
        alarm := (hours = alarm_hour) and
                 (minutes = alarm_minute) and
                 (seconds = alarm_second) ;
        cause (time_event) ;
        if alarm
          then 
            cause (alarm_event)
      until false ;
    end ; { update_time }

  begin   { main }
    setint ; { setup interupt vectors }
    abort_event := nil ;
    time_event := nil ;
    alarm_event := nil ;
    crt_region := nil ;
    time_region := nil ;
    hours := #0 ;
    minutes := #0 ;
    seconds := #0 ;
    alarm_hour := #0 ;
    alarm_minute := #0 ;
    alarm_second := #0 ;
    taskinit (#1, task_timeslice, addr(tcbtab), total_tasks) ;
    for task_sel := total_tasks downto #2 do
      begin
        activate (task_sel, task_sel, task_timeslice,
                  addr(task_stacks[task_sel]),
                  task_data_size, task_system_size) ;
        case self of
          update_time_id   : update_time ;
          display_alarm_id : display_alarm ;
          display_time_id  : display_time ;
          command_id       : command ;
          set_time_id      : 
            begin
              block(set_time_id) ;
              set_time
            end ;
          set_alarm_id     : 
            begin
              block(set_alarm_id) ;
              set_alarm
            end
        end
      end ;
    count := #1 ;
    repeat
      wait (abort_event,count,10000) ;
    until count = #0 ;
    writeln ('') ;
    writeln ('multi tasking terminated at ') ;
    write_time (hours,minutes,seconds) ;
    taskterm ;
    clearint
end.   { clock }
