module diag {$v3,0 diagnostics} ;
{
  5/17/89 : initial conversion to level 3
}
{$I/r0/token}
{$I/r0/error}
{$I/r0/varib}

  var varib external
    actual_symtab : longhex ;
    actual_memavail : longhex ;
    int_start, int_cur, ext_start, ext_cur, cur_adr,
    rel_start, rel_adr, bufstart : longhex ;
    
  procedure listout (s : string[120]) ; external ;
  
  function fromamd (r : real) : real ; entry ;
    var
      x : record
            case integer of
              1 : (rl : real) ;
              2 : (lx : longhex)
          end ;
      l1, l2, l3 : longhex register ;
    begin
      x.rl := r ;
      if r = 0
        then
          fromamd := 0
        else
          begin
            l1 := x.lx and $80000000 ; {sign}
            l2 := x.lx and $7f000000 ; {exponent}
            l3 := x.lx and $00ffffff ; {mantissa}
            l2 := l2 >> 1 ;
            if l2 >= $20000000
              then
                l2 := l2 - $01000000
              else
                l2 := l2 + $3f000000 ;
            l2 := l2 and $7f800000 ;
            l3 := l3 and $007fffff ;
            x.lx := l1 or l2 or l3 ;
            fromamd := x.rl
          end
    end ;
    
  function typeof (toke : token_type) : string [20] ; entry ;
    begin
      case toke of
        deferred  : typeof := 'deferred' ;
        boolvar   : typeof := 'boolean' ;
        charvar   : typeof := 'character' ;
        integvar  : typeof := 'integer' ;
        hexvar    : typeof := 'hex' ;
        lhexvar   : typeof := 'longhex' ;
        lintvar   : typeof := 'longinteger' ;
        realvar   : typeof := 'real' ;
        enumvar   : typeof := 'enumerated' ;
        subbool   : typeof := 'subrange boolean' ;
        subchar   : typeof := 'subrange character' ;
        subinteg  : typeof := 'subrange integer' ;
        subhex    : typeof := 'subrange hex' ;
        sublhex   : typeof := 'subrange longhex' ;
        sublint   : typeof := 'subrange longinteger' ;
        arrayvar  : typeof := 'array' ;
        stringvar : typeof := 'string' ;
        setvar    : typeof := 'set' ;
        recordvar : typeof := 'record' ;
        devicevar : typeof := 'device' ;
        valuepar  : typeof := 'value parameter' ;
        varibpar  : typeof := 'variable parameter' ;
        funcret   : typeof := 'function return'
        else        typeof := ''
      end
    end ;
    
  function storageid (st : storage_type) : string[8] ; entry ;
    begin
      case st of
        s_stack     : storageid := 'stack   ' ;
        s_external  : storageid := 'external' ;
        s_absolute  : storageid := 'absolute' ;
        s_direct    : storageid := 'direct'   ;
        s_data_ext  : storageid := 'data ext' ;
        s_data      : storageid := 'data    ' ;
        s_varib_ext : storageid := 'var. ext' ;
        s_varib     : storageid := 'varib   ' ;
        s_pcr       : storageid := 'pcr     ' ;
        s_stddev    : storageid := 'std dev ' ;
        s_nospace   : storageid := 'no space' ;
        s_recoff    : storageid := 'offset  ' ;
      end
    end ;
{
  display current lexical levels symbol table
}
  procedure sym_dump ; entry ;
    const
      stars1 = '****** ' ;
      stars2 = '******** ' ;
    var
      needhead : boolean ;
      p1 : ^symhead ;
      h : hex register ;
      
    procedure disp_sym ;
      var
        storeid : string [8] ;
        sname : string[8] ;
        idid : string [9] ;
        add : string [120] ;
        add2 : string [120] ;
        toke : token_type register ;
        p : manypt ;
      begin
        with p1^ do
          begin
            if needhead
              then
                begin
                  listout (concat(stars1,'Address  Name     Next     ',
                    'Offset   Int  Lex Storage  Id')) ;
                  needhead := false ;
                end ;
            storeid := storageid (storage) ;
            case idgroup of
              constdef    : idid := 'constant ' ;
              typedef     : idid := 'type     ' ;
              vardef      : idid := 'variable ' ;
              labeldef    : idid := 'label    ' ;
              funcdef     : idid := 'function ' ;
              procdef     : idid := 'procedure'
            end ;
            sname := substr(nameptr^,1,8) ;
            listout (concat(stars1,string(p1),' ',sname,
              substr('        ',1,9 - length(sname)),
              string(nextptr),' ',string(offset),'     ',string(intname),
              ' ', string(ord(lexlev)), 
              substr('   ',1,4-length(string(ord(lexlev)))), storeid,
              substr('   ',1,3), idid)) ;
            p.tk := p1 + align(sizeof(symhead)) ;
            toke := p.tk^ ;
            add := typeof (toke) ;
            add2 := '' ;
            case idgroup of
              constdef :
                begin
                  p.tk := p.tk + $1L ;
                  if not (toke in [boolvar, charvar, enumvar])
                    then
                      p.tk := (p.tk + $1L) and $FFFFFFFE ;
                  case toke of
                    boolvar : add2 := string(p.bl^) ;
                    charvar :
                      if p.ch^ in [' ','~']
                        then
                          add2 := concat ('''',p.ch^,'''')
                        else
                          add2 := substr(string(hex(p.ch^)),3,2) ;
                    integvar : add2 := string(p.int^) ;
                    hexvar : add2 := string(p.hx^) ;
                    lhexvar : add2 := string(p.lx^) ;
                    lintvar : add2 := string(p.li^) ;
                    realvar : add2 := string(fromamd(p.rl^), -9) ;
                    enumvar : add2 := string(ord(p.en^)) ;
                    stringvar : add2 := p.st^
                  end ;
                  add2 := concat ('value = ',add2)
                end ;
              typedef, vardef :
                begin
                  if toke in [stringvar, setvar]
                    then
                      p.tk := p.tk + $1L
                    else
                      p.tk := (p.tk + $2L) and $FFFFFFFE ;
                  case toke of
                    deferred : add2 := concat (' to ',
                                               string(p.lx^)) ;
                    hexvar :
                      with p.pt^ do
                        begin
                          case ptrf of
                            stand : add2 := typeof(base_type) ;
                            nonstd : add2 := string(base_ptr)
                            else add2 := ('nothing (normal hex)')
                          end ;
                          add2 := concat ('pointer to ',add2)
                        end ;
                    arrayvar, recordvar, devicevar :
                      add2 := concat ('with size of ',
                                      string(p.hx^));
                    stringvar, setvar :
                      add2 := concat ('with size of ',
                                      string(ord(p.ch^))) ;
                    valuepar, varibpar, funcret :
                      add2 := concat ('for procedure/function at ',
                                      string(p.lx^))
                  end
                end ;
              procdef, funcdef :
                begin
                  with p.ph^ do
                    begin
                      case comp_status of
                        proc_in_comp : add := 'in comp.' ;
                        proc_absolute : add := 'absolute' ;
                        proc_normal : add := 'normal' ;
                        proc_external : add := 'external' ;
                        proc_forward : add := 'forward' ;
                        proc_exc_in : add := 'except IC' ;
                        proc_except : add := 'exception'
                      end ;
                      add2 := concat ('Status : ',add,', Param size : ',
                                string(param_size),', Function size : ',
                                string(func_size))
                    end
                end
            end ;
            if idgroup in [funcdef, procdef]
              then
                listout (concat(stars2,add2))
              else
                if idgroup in [typedef, vardef, constdef]
                  then
                    listout (concat(stars2,'Symbol type : ',
                             add,' ',add2)) ;
          end
      end ;

    begin
      needhead := true ;
      p1 := last_sym ;
      while p1 <> nil do
        begin
          disp_sym ;        
          p1 := p1^.nextptr
        end ;
      if lex = 1
        then
          for h := $0 to $216 do
            begin
              p1 := hash_tab[h] ;
              while p1 <> nil do
                begin
                  disp_sym ;
                  p1 := p1^.nextptr
                end
            end ;
      listout(concat(stars1,'Internal : ',
          string(heapstart + internal_size - $80L - heapptr),
          ' Heap/Stack : ',string(actual_memavail),' Symtab : ',
          string(sym_bottom + actual_symtab - next_sym - $80L))) ;
      listout(concat(stars1,'Relo buffer : ',string(cur_adr - bufstart),
          ' Relo internal : ', string(int_cur - int_start) ,
          ' Relo external : ', string(ext_cur - ext_start)))
    end ;
 
Modend.
