module ngen {$v3,1 convert icode to native code} ;
{
  5/17/89 : initial conversion to version 3
  7/22/89 : add andcc and orcc
}
{$I/r0/token}
{$I/r0/code}
{$I/r0/varib}

  type
    routine_type = (
                    rpass,    {pass routine}
                    rinline,  {inline routien}
                    rdata,    {data routine}
                    rbug,     {debug routine}
                    rline,    {line number routine}
                    rpurg,    {purge symbol routine}
                    requ,     {equate routine}
                    rpspul,   {push and pull}
                    rtrans,   {tfr}
                    relatv,   {branches}
                    rcode,    {normal instructions}
                    rzap,     {stop optimizer}
                    rdcode,   {code directive}
                    rddata,   {data directive}
                    rdvarib,  {varib directive}
                    rdend,    {end directive}
                    rirq      {irq instructions}
                   ) ;
                  
    code_tab_entry = record
                       regmod : byte ;
                       routine : routine_type ;
                       name : array [#1..#4] of char
                     end ;
                    
    code_tab_type = array [code_type] of code_tab_entry ;
    
  const
    codetab = external code_tab_type ;
    
  procedure codeout ; external ;

  function typeof (toke : token_type) : string[20] ; external ;
  
  function storageid (st : storage_type) : string [8] ; external ;
  
  function labout (h : hex) : string [4] ; entry ;
    var
      s : string[4] ;
    begin
      s := string(h) ;
      while (s[0] > 1) and (s[1] = '0') do
        s := substr(s,2,3) ;
      labout := s
    end ;
      
  procedure native ; entry ;
    var
      post : byte ;
      count : byte register ;
      j : byte register ;
      opcode : code_type ;
      lth : byte register ;
      p : manypt ;
      pc : ^code_tab_entry ;
      temp : string[10] ;
      
    function labpt : string [4] ;
      var
        h : hex register ;
      begin
        h := hex(p.ch^) << 8 ;
        p.ch := p.ch + 1 ;
        h := h + hex (p.ch^) ;
        p.ch := p.ch + 1 ;
        labpt := labout (h)
      end ;
      
    function hexout (h : hex) : string [6] ;
      var
        shead : string[2] ;
      begin
        if h >= $8000
          then
            begin
              shead := '-' ;
              h := 0 - h
            end
          else
            shead := '' ;
        if h > 9
          then
            shead := concat (shead,'$') ;
        hexout := concat (shead,labout(h))
      end ;
      
    function hexpt : string [6] ;
      var
        h : hex register ;
      begin
        h := hex(p.ch^) << 8 ;
        p.ch := p.ch + 1 ;
        h := h + hex (p.ch^) ;
        p.ch := p.ch + 1 ;
        hexpt := hexout (h)
      end ;
    
    function longpt : string [10] ;
      var
        mix : record
                case boolean of
                  false : (b1, b2, b3, b4 : byte) ;
                  true  : (l : longhex)
              end ;
        shead : string[2] ;
        s : string[8] ;
      begin
        mix.b1 := p.ch^ ;
        p.ch := p.ch + 1 ;
        mix.b2 := p.ch^ ;
        p.ch := p.ch + 1 ;
        mix.b3 := p.ch^ ;
        p.ch := p.ch + 1 ;
        mix.b4 := p.ch^ ;
        p.ch := p.ch + 1 ;
        if mix.l >= $80000000
          then
            begin
              shead := '-' ;
              mix.l := 0 - mix.l
            end
          else
            shead := '' ;
        if mix.l > 9
          then
            shead := concat (shead,'$') ;
        s := string(mix.l) ;
        while (s[0] > 1) and (s[1] = '0') do
          s := substr(s,2,7) ;
        longpt := concat(shead,s)
      end ;
      
    procedure opout ;
      var
        i : byte register ;
      begin
        output_buf := ' ' ;
        for i := 1 to 4 do
          if pc^.name[i] = ' '
            then
              exit
            else
              output_buf := concat (output_buf,pc^.name[i]) ;
        if (output_buf[2] = 'a') and (output_buf = ' andc')
          then
            append (output_buf, 'c') ;
        append (output_buf, ' ')
      end ;
      
    function regout : string[4] ;
      begin
        case post and $18 of
          0 : regout := ',x' ;
          8 : regout := ',y' ;
          $10 : regout := ',u' ;
          $18 : regout := ',pcr'
        end
      end ;

    begin  {native}
      output_buf := '' ;
      lth := icode_buf[0] - 1 ; {minus opcode}
      opcode := enum(icode_buf[1]) ;
      post := icode_buf[2] ;
      p.ch := addr(icode_buf[3]) ;
      pc := addr(codetab[opcode]) ;
      case pc^.routine of
        rpass, rinline : 
          begin
            output_buf := substr(icode_buf,2,ord(lth)) ;
            codeout            
          end ;            
        rdata : 
          begin
            count := 0 ; {number on line so far}
            p.hx := addr(icode_buf[2]) ;
            for j := 1 to lth do
              begin
                if count = 0
                  then
                    output_buf := ' fcb '
                  else
                    output_buf := concat (output_buf,',') ;
                output_buf := concat (output_buf,hexout(hex(p.ch^))) ;
                p.ch := p.ch + 1 ;
                count := succ (count) ;
                if count > 8
                  then
                    begin
                      count := 0 ;
                      codeout
                    end
              end ;
            if count <> 0
              then
                codeout ;
          end ;
        rbug :
          begin
            case icode_buf[2] of
              0 : {procedure start}
                begin
                  output_buf := concat('* proc start, block=',hexpt,
                                  ' param size=',hexpt,' name=',p.st^)
                end ;
              1 : {external procedure}
                begin
                  output_buf := concat('* ext proc, block=',hexpt,
                                  ' param size=',hexpt,' name=',p.st^)
                end ;
              2 : {end of procedure}
                begin
                  output_buf := concat('* proc end, block=',hexpt,
                                       ' name=',p.st^)
                end ;
              3 : {type/variable declaration}
                begin
                  output_buf := concat('type/varib ',p.st^) ;
                  p.ch := p.st + longhex(p.ch^ + 1) ;
                  output_buf := concat(output_buf,' ', longpt,
                                       ' ',hexpt) ;
                  case p.en^ of
                    typedef  : output_buf := 
                                 concat(output_buf, ' typedef ') ;
                    vardef   : output_buf := 
                                 concat(output_buf, ' vardef ')
                  end ;
                  p.en := p.en + 1 ;
                  output_buf := concat(output_buf,
                                       storageid(p.en^),' ') ;
                  p.en := p.en + 1 ;
                  output_buf := concat('* ',output_buf, hexpt,' ',
                                                      typeof(p.en^)) ;
                  if p.en^ in [valuepar .. funcret]
                    then
                      begin
                        p.en := p.en + 1 ;
                        output_buf := concat(output_buf,
                                             ' ',typeof(p.en^))
                      end ;
                  if p.en^ = deferred
                    then
                      begin
                        p.en := p.en + 1 ;
                        output_buf := concat(output_buf, ' ',longpt)
                      end 
                end ;
              4 : {pointer update}
                output_buf := concat('* pointer update ',
                                     longpt, ' ', longpt) ;
              5 : {module name}
                output_buf := concat('* module ',p.st^) ;
              6 : {program name}
                output_buf := concat('* program ',p.st^) ;
              7 : {lex}
                output_buf := concat('* lex ',string(ord(p.ch^)))
            end ;
            codeout
          end ; {bug}
        rline :
          begin
            output_buf := concat('* line number ',string(integer
                            (hex(icode_buf[2]) << 8 +
                             hex(icode_buf[3])))) ;
            codeout
          end ;
        rpurg  :
          if debug_opt and debug_en
            then
              begin
                p.ch := addr(icode_buf[2]) ;
                output_buf := concat('* purge label ',labpt) ;
                codeout
              end ;
        rzap   :
          if debug_opt and debug_en
            then
              begin
                output_buf := '* ZAP' ;
                codeout
              end ;
        requ   : 
          begin
            case post of
              p_imm+p_eqlp : {lbxxxx equ *}
                output_buf := concat ('.$',labpt,' equ *') ;
              p_imm+p_eqep : {xxxx equ *}
                output_buf := concat (p.st^,' equ *') ;
              else           {xxxxx equ $yyyy}
                begin
                  output_buf := concat (p.st^,' equ ') ;
                  p.hx := p.st + longhex(p.ch^) + 1 ;
                  output_buf := concat (output_buf,hexpt)
                end
            end ;
            codeout
          end ;
        rpspul :
          begin
            opout ;
            if (post and p_ips) <> 0
              then
                begin
                  append (output_buf, 's') ;
                  if post and $f <> 0
                    then
                      append (output_buf, ',')
                end ;
            if (post and p_ipy) <> 0
              then
                begin
                  append (output_buf, 'y') ;
                  if post and 7 <> 0
                    then
                      append (output_buf, ',')
                end ;
            if (post and p_ipx) <> 0
              then
                begin
                  append (output_buf, 'x') ;
                  if post and 3 <> 0
                    then
                      append (output_buf, ',')
                end ;
            if (post and p_ipd) = p_ipd
              then
                append (output_buf, 'd')
              else if (post and p_ipb) <> 0 
                then
                  append (output_buf, 'b')
              else if (post and p_ipa) <> 0
                then
                  append (output_buf, 'a') ;
            codeout
          end ;
        rtrans :
          begin
            opout ;
            case post of
              p_itba : append (output_buf, 'b,a') ;
              p_itab : append (output_buf, 'a,b') ;
              p_itdx : append (output_buf, 'd,x') ;
              p_itxd : append (output_buf, 'x,d')
            end ;
            codeout
          end ;
        rcode :
          begin
            opout ;
            if lth <> 0 {not inherent}
              then
                if (post and p_idx) <> 0
                  then
                    begin {indexed}
                      if (post and p_idxi) <> 0
                        then
                          append (output_buf, '[') ;
                      case post and 7 of
                        p_ixfx : output_buf := concat(output_buf,hexpt,regout) ;
                        p_ixex :
                          begin
                            output_buf := concat(output_buf, p.st^) ;
                            p.st := p.st + longhex(p.ch^ + 1) ;
                            output_buf := concat(output_buf, regout)
                          end ;
                        p_ixb : output_buf := concat(output_buf,'b',regout) ;
                        p_ixd : output_buf := concat(output_buf,'d',regout) ;
                        p_ixp : output_buf := concat(output_buf,'0',regout,'+') ;
                        p_ixpp : output_buf := concat(output_buf,'0',regout,'++') ;
                        p_ixm : output_buf := concat(output_buf,'0,-',substr(regout,2,3))
                        else  output_buf := concat(output_buf,'0',regout)
                      end ;
                      if (post and p_idxi) <> 0
                        then
                          append (output_buf, ']')
                    end
                  else
                    begin
                      if (post and p_abs) = 0
                        then
                          begin
                            append (output_buf, '#') ;
                            if opcode in [c_addb,c_anda..c_andcc,c_cmpb,c_eora,c_eorb,
                                          c_lda,c_ldb,c_ora..c_orcc,c_subb]
                              then
                                output_buf := concat(output_buf, hexout(hex(p.ch^)))
                              else
                                output_buf := concat(output_buf, hexpt)
                          end
                        else
                          begin
                            if (post and p_dir) <> 0
                              then
                                append (output_buf, '<') ;
                            output_buf := concat(output_buf,hexpt)
                          end
                    end ;
            codeout
          end ;
        relatv : 
          begin
            if (post and p_rell) <> 0
              then
                output_buf := ' l'
              else
                output_buf := ' ' ;
            case post and $f of
              p_bra : temp := 'bra' ;
              p_bsr : temp := 'bsr' ;
              p_bhi : temp := 'bhi' ;
              p_bls : temp := 'bls' ;
              p_bcc : temp := 'bhs' ;
              p_bcs : temp := 'blo' ;
              p_bne : temp := 'bne' ;
              p_beq : temp := 'beq' ;
              p_bvc : temp := 'bvc' ;
              p_bvs : temp := 'bvs' ;
              p_bpl : temp := 'bpl' ;
              p_bmi : temp := 'bmi' ;
              p_bge : temp := 'bge' ;
              p_blt : temp := 'blt' ;
              p_bgt : temp := 'bgt' ;
              p_ble : temp := 'ble'
            end ;
            append (output_buf, temp, ' ') ;
            if (post and p_reex) <> 0 { external }
              then
                output_buf := concat (output_buf,p.st^)
              else
                output_buf := concat (output_buf,'.$',labpt) ;
            codeout
          end ;
        rdcode :
          begin
            output_buf := ' code' ;
            codeout
          end ;
        rddata :
          begin
            output_buf := ' data' ;
            codeout
          end ;
        rdvarib : 
          begin
            output_buf := ' varib' ;
            codeout
          end ;
        rdend :
          begin
            output_buf := ' end' ;
            codeout
          end ;
        rirq :
          begin
            case post of
              p_save : output_buf := ' pshs u,y,x,d' ;
              p_ldy : output_buf := ' ldy global' ;
              p_rest : output_buf := ' puls u,y,x,d' ;
              p_rti : output_buf := ' rti' ;
              p_leay : output_buf := ' leay 10,u' ;
            end ;
            codeout
          end
      end
    end ;

modend.
